import logging

import lab.tensorflow as B
import tensorflow as tf

from ..minimise import make_l_bfgs_b, make_adam, exception

__all__ = ["minimise_l_bfgs_b", "minimise_adam"]

log = logging.getLogger(__name__)


def _wrap_f(vs, names, f, jit):
    # Differentiable assignments will overwrite the variables, so make a copy.
    vs_copy = vs.copy()

    # Keep track of function evaluations.
    f_evals = []

    def f_vectorised(x):
        vs_copy.set_vector(x, *names, differentiable=True)
        return f(vs_copy)

    if jit:
        f_vectorised = tf.function(f_vectorised, autograph=False)

    def f_wrapped(x):
        x_tf = B.cast(vs.dtype, x)

        # Compute objective function value and gradient.
        try:
            with tf.GradientTape() as t:
                t.watch(x_tf)
                obj_value = f_vectorised(x_tf)
                grad = t.gradient(obj_value, x_tf, unconnected_gradients="zero")
        except Exception as e:
            return exception(x, e)

        # Convert to NumPy.
        obj_value, grad = B.to_numpy(obj_value, grad)

        f_evals.append(obj_value)
        return obj_value, grad

    return f_evals, f_wrapped


minimise_l_bfgs_b = make_l_bfgs_b(_wrap_f)
minimise_adam = make_adam(_wrap_f)



from itertools import product

import lab as B
import numpy as np
from varz import Vars, Initialiser, Packer

from .util import allclose, vs


def test_initialiser(vs):
    init = Initialiser()

    # Initialise two variables.
    vs.get(1.0, name="a")
    vs.pos(2.0, name="b")

    # Define some initialisations.
    init.assign("a", [-3.0, 4.0])
    init.assign("b", [5.0, 6.0])
    inits = init.generate(vs)

    # Test the initialisations.
    for initialiser, values in zip(inits, product([-3.0, 4.0], [5.0, 6.0])):
        initialiser()
        assert vs["a"] == values[0]
        allclose(vs["b"], values[1])


def test_packer():
    a, b, c = B.randn(5, 10), B.randn(20), B.randn(5, 1, 15)

    for packer, args in zip(
        [Packer(a, b, c), Packer([a, b, c])], [(a, b, c), ((a, b, c),)]
    ):
        # Test packing.
        packed = packer.pack(*args)
        assert B.rank(packed) == 1

        # Test unpacking.
        a_, b_, c_ = packer.unpack(packed)
        allclose(a, a_)
        allclose(b, b_)
        allclose(c, c_)



from setuptools import find_packages, setup

requirements = ["numpy>=1.16", "scipy>=1.3", "plum-dispatch", "backends>=0.5.0", "wbml"]

setup(
    packages=find_packages(exclude=["docs"]),
    python_requires=">=3.6",
    install_requires=requirements,
    include_package_data=True,
)
